C Ramsey3d.for:
C
C Purpose: Sovle the stochastic growth model with non-negative investment using 
C          Projection methods


C Module definitions:

C Parameters of the model and of the algorithm
      Module Def1
	
	Real(8) alpha, beta, eta, delta, rho, sigma
	Real(8) z0, k0, zmin, zmax, kmin, kmax, ks
	Real(8) bounds1(1:2,1:2), bounds2(1:2,1:2)      
      Real(8), allocatable :: gvec(:), zgrid(:), pmat(:,:)
      Real(8), allocatable :: ct(:), kt(:), zt(:), mt(:)

      Integer(4) degree, degree1, degree2, ng			

C Parameters of the algorithm     
	Logical mca, complete 
	Integer(4) knodes, znodes, nz, nobs, nofs, ta, lags
	Real(8)  size
	
	End Module Def1

C Module for possible Errors that need to be known by various subroutines
      Module Errors

	Integer(2) FError, ErrorType

	End Module Errors


C Here begins the main program
      program Ramsey3d

      use DFLIB
	use IMSLF90	
      use Def1
	use Errors
	use MNR
C	use QNP
	      
      Integer(4) i, j, hits(2,2), freq(1:3), tmax	
	Real(4) Asec
      Real(8), Allocatable :: Gamma1(:), Gamma2(:), Gamma3(:)
	Real(8) StDev, dm, chi(2), fit, crit(1:6), stat(3)	
	Real(8) xscale(6), fscale, rparam(7)
	Integer Iparam(7)

	Character(256) Laufzeit
	Character(50) MyDate

	Logical lsq, initialize, ok, InitialSettings

      
	External f1, f2, f3, sys1, sys2, sys3

	Type (rccoord) curpos		

C Read Parameters from input file and compute derived parameters
C Parameters of the algorithm
      Call GetL('lsq','Parameters.txt',lsq)
	Call GetL('mca','Parameters.txt',mca)
	Call GetL('complete','Parameters.txt',complete)
	Call GetL('initialize','Parameters.txt',initialize)
      Call GetI('knodes','Parameters.txt',knodes)
	Call GetI('znodes','Parameters.txt',znodes)
	if (complete) then
	   Call GetI('degree','Parameters.txt',degree)
	   ng=(degree+1)*(degree+2)/2	   
	   degree1=degree
	   degree2=degree
	else
	   Call GetI('degree1','Parameters.txt',degree1)
	   Call GetI('degree2','Parameters.txt',degree2)
	   ng=degree1*degree2
	endif
      Call GetI('nobs','Parameters.txt',nobs)
	Call GetI('nofs','Parameters.txt',nofs)
	Call GetI('ta','Parameters.txt',ta)
	Call GetI('lags','Parameters.txt',lags)
      Call GetD('size','Parameters.txt',size)
	Call GetI('nz','Parameters.txt',nz)

C Parameters of the model
	Call GetD('alpha' ,'Parameters.txt',alpha)
	Call GetD('beta'  ,'Parameters.txt',beta )
	Call GetD('eta'  ,'Parameters.txt',eta ) 
	Call GetD('delta','Parameters.txt',delta)	
	Call GetD('rho'  ,'Parameters.txt',rho  )
	Call GetD('sigma','Parameters.txt',sigma)
      
C Allocate space
      tmax=nobs+ta+lags+1
      Allocate (gvec(1:ng),gamma1(1:ng),gamma2(1:ng),gamma3(1:ng),
     + ct(1:tmax), zt(1:tmax), kt(1:tmax+1),
     + mt(1:tmax))

C Prepare output window
      ok=InitialSettings()
      Call MakeSmallWindow(10,"Program Progress",2)

C Fix starting time
      Asec=Secnds(0.0)
      
      open(11,file='LogFile.txt')	

	WRITE(11,"(A50)") MyDate()		

C Compute Boundaries
      zmin=-size*(sigma/dsqrt(1.-(rho**2))) ! lower bound for productivity shock
	zmax=-zmin                            ! upper bound for producitvity shock
      ks=((alpha*beta)/(1.-beta*(1.-delta)))**(1./(1.-alpha)) ! deterministic steady state for the stock of capital    
	kmin=0.2*ks  ! lower bound for capital stock	
	kmax=2.0*ks  ! upper bound for capital stock
      bounds1(1,1)=zmin
	bounds1(1,2)=zmax
	bounds1(2,1)=dlog(kmin)
	bounds1(2,2)=dlog(kmax)

      bounds2(1,1)=zmin
	bounds2(1,2)=zmax
	bounds2(2,1)=kmin
	bounds2(2,2)=kmax

C Compute grid for productivity shock
      if (mca) then
	   Allocate(zgrid(1:nz),pmat(1:nz,1:nz))	   
	   Call MarkovAR(size,nz,rho,sigma,zgrid,pmat)
	endif


C load initial value
      if (.not. initialize) then
	  open(20,file='ParIni.txt')
	  read(20,*) gamma1
	  close(20)
	endif


      if (initialize) then
	  if (lsq) then
          Call GSearch1(F2,ng,Gamma1,Fit)	  	
	  else
	    Call GSearch1(F3,ng,Gamma1,Fit)
	endif
        Call ClearScreen($GCLEARSCREEN)        
	  CALL SETTEXTPOSITION(1,2,curpos)
        Write(10,"(A,F15.10)") 'Function value of best solution:', fit
	  Write(11,"(A)") 'Solution from Search1:'
        if (complete) then
	      Do I=1,ng
		    write(11,"(F10.6)") Gamma1(i)
	      End Do
	  else
	    Do i=1,degree1
	       Do j=1,degree2
	          Write(10,"(5F10.6)") Gamma1(degree1*(i-1)+j)
	       End Do
	    End Do
	  Endif
        Write(11,"(A,F15.10)") 'Function value of best solution:', fit
      Endif

C Refine solution using either QuasiNewton or FixvMN

C Use ISLM routine DNEQNF
C     crit(1)=1.E-7 !stopping criterion
C     i=100         !number of iterations
C     Call DNEQNF(Sys3,crit(1),ng,i,Gamma1,Gamma2,fit)
C      Jac=1

      if (lsq) then
         Call QuasiNewton1(F1,ng,Gamma1,Gamma2,crit)
	else
	   if (complete) then
	      Call FixvMN(ng,Gamma1,Sys1,Gamma2,crit)
	   else
	      Call FixvMN(ng,Gamma1,Sys2,Gamma2,crit)
	   endif
	endif	

C Output solution
      if (crit(1) .eq. 0.0) then
	   Call ClearScreen($GCLEARSCREEN)        
    	   CALL SETTEXTPOSITION(1,2,curpos)
	   if (lsq) then
            Write(10,"(A)") 'Normal Termination from QuasiNewton'
	      Write(11,"(A)") 'Normal Termination from QuasiNewton'
	   else
	      Write(10,"(A)") 'Normal Termination from FixvMN'
	      Write(11,"(A)") 'Normal Termination from FixvMN'
	   endif
	Endif
      if (lsq) then
	  if (crit(1) .eq. 2.0) then
          Call ClearScreen($GCLEARSCREEN)        
    	    CALL SETTEXTPOSITION(1,2,curpos)
        Write(10,"(A)") 'No further decrease of function value possible'
	  Write(11,"(A)") 'No further decrease of function value possible'
	  elseif (crit(1) .gt. 2.0) then
	   Call ClearScreen($GCLEARSCREEN)        
    	   CALL SETTEXTPOSITION(1,2,curpos)         
	   WRITE(10,"(A,I1)") 'Termination code from QN', IDINT(crit(1))
	   WRITE(11,"(A,I1)") 'Termination code from QN', IDINT(crit(1))
	  endif
	endif	
	
	if (crit(1) .ne. 0.0) then
	  if (.not. lsq) then
          Call ClearScreen($GCLEARSCREEN)        
    	    CALL SETTEXTPOSITION(1,2,curpos)
          Write(10,"(A)") MNR_Message(IDINT(crit(1)))
	    Write(11,"(A)") MNR_Message(IDINT(crit(1)))
	  endif
	endif
	if (lsq) then	
	    Write(10,"(A,F15.8)") 'Max(RelGrad)       = ', crit(2)
	    Write(11,"(A,F15.8)") 'Max(RelGrad)       = ', crit(2)
	    Write(10,"(A,F15.8)") 'Function Value     = ', crit(3)
	    Write(11,"(A,F15.8)") 'Function Value     = ', crit(3)
	    Write(10,"(A,I5)")  'Iterations         = ', IDINT(crit(4))
	    Write(11,"(A,I5)")  'Iterations         = ', IDINT(crit(4))	   	   
	else
	    Write(10,"(A,F15.8)") 'Max(abs(fx))       = ', crit(2)
	    Write(11,"(A,F15.8)") 'Max(abs(fx))       = ', crit(2)
	    Write(10,"(A,F15.8)") 'Max(|dx/x|)        = ', crit(3)
	    Write(11,"(A,F15.8)") 'Max(|dx/x|)        = ', crit(3)
	    Write(10,"(A,F15.8)") 'Max. Relgrad       = ', crit(4)
	    Write(11,"(A,F15.8)") 'Max. Relgrad       = ', crit(4)
	    Write(10,"(A,F15.8)") '0.5*fx*fx          = ', crit(5)
	    Write(11,"(A,F15.8)") '0.5*fx*fx          = ', crit(5)	    
	    Write(10,"(A,I5)")  'Iterations         = ', IDINT(crit(6))
	    Write(11,"(A,I5)")  'Iterations         = ', IDINT(crit(6))	   	   
	endif
	Write(11,"(A)") 'Solution:'
      if (complete) then
	    Do i=1,ng	   
	       Write(11,"(F10.6)") Gamma2(i)
	    End Do
	  else
	    Do i=1,degree1
	              Write(11,"(5F10.6)")
     +              (Gamma2(degree2*(i-1)+j),j=1,degree2)
	    End Do
	Endif
      

      Call ElapsedTime(Secnds(Asec),Laufzeit)      
      Call GetTextPosition(curpos)
	Call SetTextPosition(curpos.row+1,curpos.col,curpos)
	Write(10,"('Run time=',A)") Laufzeit
	Write(11,"('Run time=',A)") Laufzeit
      if (crit(1) .eq. 0.0) then
	   Call GetTextPosition(curpos)
	   Call SetTextPosition(curpos.row+1,2,curpos)
	   Write(10,"(A)") 'Simulate the model. Please wait'
	   Asec=Secnds(0.0)
	   Call Simulate(Gamma2,freq,hits,stat)
	   Write(11,"('Total valid Simulations:',I5)") freq(3)
	   Write(11,"('in the lower tail      :',I5)") freq(1)
	   Write(11,"('in the upper tail      :',I5)") freq(2)
	   Write(11,"('Percentage in the lower tail:',F6.2)")
     +     100*dble(freq(1))/dble(freq(3))
	   Write(11,"('Percentage in the upper tail:',F6.2)")
     +     100*dble(freq(2))/dble(freq(3))
	   Write(11,"('z<zmin: ',I3)") hits(1,1)
	   Write(11,"('z>zmax: ',I3)") hits(1,2)
         Write(11,"('k<kmin: ',I3)") hits(2,1)
	   Write(11,"('k>kmax: ',I3)") hits(2,2)
         Write(11,"('Standard dev. of consumption= ',F15.10)") stat(1)
	   Write(11,"('Standard dev. of capital    = ',F15.10)") stat(2)
	   Write(11,"('Stardard dev. of shock 	   = ',F15.10)") stat(3)
	   Call ElapsedTime(Secnds(Asec),Laufzeit)      
      endif
	
         Write(11,"('Parameters of the model:')")
	   Write(11,"('alpha=  ',F7.3)") alpha
	   Write(11,"('beta =  ',F7.3)") beta
	   Write(11,"('eta  =  ',F7.3)") eta
	   Write(11,"('delta=  ',F7.3)") delta
	   Write(11,"('rho  =  ',F7.3)") rho
	   Write(11,"('sigma=  ',F7.4)") sigma
	   Write(11,"('Parameters of the simulations:')")
	   Write(11,"('complete =  ',L1)") complete
	   Write(11,"('mca      =  ',L1)") mca
	   Write(11,"('size     =  ',F7.3)") size
	   Write(11,"('nobs=  ',I6)") nobs
	   Write(11,"('ta  =  ',I6)") ta
         Write(11,"('lags=  ',I6)") lags	   
	   Write(11,"('nofs=  ',I8)") nofs
	   if (lsq) then
	     Write(11,"('knodes=',I4)") knodes
	     Write(11,"('znodes=',I4)") znodes
	   endif
	   
	close(11)
	close(15)	         

	End


C ********************************************  Subroutines *********************************************************

C f1(npar,gvec): the minimizer of this function is the least squares approximation of the conditional expectations function
C

      Real(8) Function f1(npar,gamma)

      use Def1
	use Errors

      Integer(4) Npar
	Real(8) gamma(npar)

      Integer(4) n(2)
      Real(8) RSQ, GC_Int2

	External RSQ

      gvec=gamma
   
      n(1)=znodes
	n(2)=knodes
      
      FError=0 ! clear Error flag 

      f1=GC_Int2(RSQ,bounds2,n)

      If (IsNaN(f1)) FError=1

      Return

      End Function

C f2(npar,gvec): the minimizer of this function is the least squares approximation of the conditional expectations function
C                To be consistent with the calling conventions of Search1, this subroutine returns
C                -1, if it is not possible to evaluate f at gvec
C
C           From the main program: zmin,zmax, kmin, and kmax must be stored in bounds

      Real(8) Function F2(npar,gamma)

      use Def1
	use Errors

      Integer(4) Npar
	Real(8) gamma(npar)

      Integer(4) n(2)
      Real(8) RSQ, GC_Int2

	External RSQ

      gvec=gamma
   
      n(1)=znodes
	n(2)=knodes

      FError=0 ! clear Error flag 

      f2=GC_Int2(RSQ,bounds2,n)

      If (IsNaN(F2)) F2=-1.0

      Return

      End Function

C F3 this function is used to find starting values for the collocation method. It just calls
C    Sys1 or Sys2 and returns the maximum absolute value of fx

      Real(8) Function F3(n,Gamma)

	use Def1
      use Errors

	Integer(4) n
	Real(8) Gamma(n)

      Real(8) fx(n)

	FError=0 
      
	if (complete) then
	   Call Sys1(n,n,Gamma,fx)
	else
	   Call Sys2(n,n,Gamma,fx)
	endif

      if (FError .ne. 0) then
	   F3=-1.0
	else
         F3=MaxVal(dabs(fx))
	endif

	return

	End Function

C RSQ(z,k) returns the squared residual as given by Resid

      Real(8) Function RSQ(z,k)      

	Real(8) z, k, Resid

	External Resid

      RSQ=Resid(z,k)**2

      Return

	End Function

C Sys1(gvec): defines the rhs of the system of non-linear equations whose zero is the
C             collocation solution of the model. Requires a complete polynomial

      Subroutine Sys1(m,n,par,fx)

      use Def1
	use IMSLF90	

	Integer(4) m, n
	Real(8) par(n), fx(m)


      Integer(4) i, j, p1, p2, k
	Real(8) zbar(degree+1),kbar(degree+1), pi, Dconst, Resid

	External Resid

	pi=Dconst('pi')    
  
      p1=degree+1
	
      Do i=1,p1
	   zbar(i)=dcos(((dble(2.0*i)-1.0)/(dble(2.*p1))) *pi)
	   zbar(i)=0.5*(zbar(i)+1.0)*(zmax-zmin) + zmin
	End Do
	
	Do i=1,p1
	   kbar(i)=dcos(((dble(2.0*i)-1.0)/(dble(2.*p1)))*pi)   
	   kbar(i)=0.5*(kbar(i)+1.0)*(kmax-kmin) + kmin
	End Do

	gvec=par      

      p1=degree+1
	p2=degree+2
	k=1
      Do i=1,p1
            Do j=1,(p2-i)
              fx(k)=Resid(zbar(j),kbar(i))
              k=k+1      
            End Do
      End Do
      

      Return

	End Subroutine

	Subroutine Sys3(x,fx,n)

	Integer(4) n
	Real(8) x(n), fx(n)

	External Sys1

	Call Sys1(n,n,x,fx)

	Return

	End Subroutine


C Sys2(gvec): defines the rhs of the system of non-linear equations whose zero is the
C             collocation solution of the model. Requires a product base polynomial

      Subroutine Sys2(m,n,par,fx)

      use Def1
	use IMSLF90	

	Integer(4) m, n
	Real(8) par(n), fx(m)


      Integer(4) i, j
	Real(8) zbar(degree1),kbar(degree2), pi, Dconst, Resid

	External Resid

	pi=Dconst('pi')    
           
      Do i=1,degree1
	   zbar(i)=cos(((dble(2.0*i)-1.0)/(dble(2.*degree1))) *pi)
	   zbar(i)=0.5*(zbar(i)+1.0)*(zmax-zmin) + zmin
	End Do
	
	Do i=1,degree2
	   kbar(i)=cos(((dble(2.0*i)-1.0)/(dble(2.*degree2)))*pi)   
	   kbar(i)=0.5*(kbar(i)+1.0)*(kmax-kmin) + kmin
	End Do

	gvec=par      

     
      Do i=1,degree1
	      Do j=1,degree2
	         fx(degree2*(i-1)+j)=Resid(zbar(i),kbar(j))
	      End Do
      End Do
   
      Return

	End Subroutine

C Resid(z,k) returns the residual at the point (z,k), the current vector of
C        parameters gvec must be given.

      Real(8) Function Resid(z,k)

      use Def1      

	Real(8) z, k

      Real(8) psi,  phi, GetPsi1, GetPsi2, Phi1, Phi2
   
      z0=z
      k0=k

	if (complete) then
         psi=GetPsi1(gvec,z,k)
	else
	   psi=GetPsi2(gvec,z,k)
	endif

	if (IsNan(psi)) then
	  Resid=psi
	  Return
	endif

      if (mca) then
	   phi=Phi1(z,k)
	else
	   phi=Phi2(z,k)
	endif
	
	if (IsNan(phi)) then
	   Resid=phi
	   Return
	endif
	
      Resid=psi-phi

	Return

	End Function


C GetPsi1 returns Psi(Gamma,z,k) in the case of a complete polynomial

      Real(8) Function GetPsi1(gamma,z,k)

	use Def1
	use Errors

      Real(8) gamma((degree+1)*(degree+2)/2), z, k
	
	Real(8) x(1:2), ChebEval2

	x(1)=z
	x(2)=dlog(k)
      FError=0
	GetPsi1=ChebEval2(gamma,degree,x,bounds1)      

      If (GetPsi1 .gt. 700.) then      
	   FError=1
	   GetPsi1=0./0.
	   ErrorType=1
	   return
	else
	   GetPsi1=beta*dexp(GetPsi1)
	   if (GetPsi1 .le. 0.0) then
	      FError=1
	      GetPsi1=0./0.
	      ErrorType=2
	   endif
	endif      

	Return
	End Function

C GetPsi2 returns Psi(Gamma,z,k) in the case of a product base polynomial
C
      Real(8) Function GetPsi2(gamma,z,k)

	use Def1
	use Errors

      Real(8) gamma(degree1*degree2), z, k
      Real(8) gmat(degree1,degree2)	
	
	Integer(4) n(2), i, j
	Real(8) x(1:2), ChebEval3

      x(1)=z
	x(2)=dlog(k)

      n(1)=degree1
	n(2)=degree2

      Do i=1,degree1,1
	   Do j=1, degree2,1
            gmat(i,j)=gamma(degree2*(i-1)+j)
         End Do
	End Do

	FError=0
	GetPsi2=ChebEval3(gmat,n,x,bounds1)

      If (GetPsi2 .gt. 700.) then	  
	   FError=1
	   ErrorType=1
	   GetPsi2=0./0.
	   return
	else
	   GetPsi2=beta*dexp(GetPsi2)
	   if (GetPsi2 .le. 0.0) then
	      FError=1
	      ErrorType=2
	      GetPsi2=0./0
	   endif
 	endif

	Return
	End Function


C Phi1: returns the conditional expectations for given z and k. This is computed
C          from the Markov chain approximation of the stochastic AR(1) process
C          of the technology shock. 


      Real(8) Function Phi1(z,k)

	use Def1	

      Real(8) z, k
 
      Integer(4) i, j
	Real(8) x, y, p(nz), PhiZ

      z0=z
      k0=k

      Call ARMC1(z0,p)

      x=0
      
      Do i=1,nz
         y=PhiZ(zgrid(i))
         if (IsNaN(y)) then
	      Phi1=y
		  Return
 		Endif		
         x=x+y*p(i)
      End Do

      Phi1=beta*x

	Return

	End Function


C Phi2: Returns the conditional expectation for given z and k.
C          In this procedure, this expression is computed from
C          the integral of p(eps)*(u'(C1)*(1-delta+alhpa*k1^(alpha-1))-(1-delta)*mu1)),
C          where p(eps) is given from the  probability density function
C          of the respective normal distribution 


      Real(8) Function Phi2(z,k)

      use Def1	

      Real(8) z, k

	Real(8) phi, PhiEps, GC_Int1
   
      External PhiEps

      z0=z
	k0=k

      phi=GC_Int1(PhiEps,bounds1(1,1:2),100)
	
	if (IsNan(phi)) then
	   Phi2=phi
	   Return
	Endif
	
      Phi2=beta*phi

	Return

	End Function

C PhiZ: returns the u'(c')(1-delta+f'(k')-(1-delta)*mu', which is involved in the
C       computation of phi(z,k). The program assumes gvec, k0 and z0 given from the
C       main program.


      Real(8) Function PhiZ(z)

	use Def1
	use Errors

      Real(8) z

      Real(8) c0, c1, k1, k2, m1, x, psi0, psi1, GetPsi1, GetPsi2

	if (complete) then
          psi0=GetPsi1(gvec,z0,k0)
	else
	    psi0=GetPsi2(gvec,z0,k0)
	endif

      if (IsNaN(psi0)) then
	    PhiZ=psi0
		Return
	Endif

C First Step: Compute k1 
      c0=psi0**(-1./eta)
      if (c0 .lt. 0.0) then
	 FError=1
	 ErrorType=3
	 PhiZ=0./0.
	 Return
	endif

      k1=dexp(z0)*(k0**alpha)+(1-delta)*k0-c0
      if (k1 .lt. (1-delta)*k0) then
	   k1=(1-delta)*k0
	Endif

C Second Step: Compute c1
      if (complete) then
         psi1=GetPsi1(gvec,z,k1)
	else
	   psi1=GetPsi2(gvec,z,k1)
	endif

      if (IsNan(psi1)) then
	  PhiZ=psi1
	  Return
	Endif
	      
      c1=psi1**(-1/eta)
      if (c1 .lt. 0.0) then
	 FError=1
	 ErrorType=4
	 PhiZ=0./0.
	 Return
	Endif
	
      k2=dexp(z)*(k1**alpha)+(1-delta)*k1-c1
      
      if (k2 .lt. (1-delta)*k1) then
         c1=dexp(z)*(k1**alpha)
         m1=c1**(-eta)-psi1
	   if (m1 .lt. 0.0) then
	     FError=1
	     ErrorType=5
		 PhiZ=0./0.
		 Return
	   Endif
      else
         m1=0
      endif

      x=(c1**(-eta))*(1-delta + alpha*dexp(z)*(k1**(alpha-1)))
     +  -(1-delta)*m1
    
      PhiZ=x

	Return

	End Function


C PhiEps: returns the integrand in the definition of the conditional expectation on the
C         rhs of the model's Euler equation as a function of epsilion in the transition
C         equation of ln Z:= z, z'=rho*z + epsilon. This is used to compute the
C         conditional expectations via Gauss-Chebyshev quadrature.
C
C         The program assumes the vector of coefficients of the expectations function psi, gvec,
C         as given in the main program, as well as the current values of ln z and the capital stock
C         K. We store those two in z0 and k0.
C         Also, alpha, beta, delta, eta, rho, and sigma are taken from the main program.

      Real(8) Function PhiEps(Eps)

      use Def1
	use Errors      
      use IMSLF90	
	Real(8) Eps

      Real(8) c0, c1, k1, k2, m1, x, p, psi0, psi1, GetPsi1, GetPsi2, pi
	Real(8) dconst

	pi=dconst('pi')

      if (complete) then
        psi0=GetPsi1(gvec,z0,k0)
	else
	  psi0=GetPsi2(gvec,z0,k0)
	endif

      if (IsNan(psi0)) then
	   PhiEps=psi0
	   Return
	Endif

C First Step: Compute k1 
      c0=psi0**(-1/eta)
      if (c0 .lt. 0.0) then
	 FError=1
	 ErrorType=3
	 PhiEps=0./0.
	 Return
	Endif
      
      k1=dexp(z0)*(k0**alpha)+(1-delta)*k0-c0
      
      if (k1 .lt. (1-delta)*k0) then
	   k1=(1-delta)*k0
	Endif

C Second Step: Compute c1 
      if (complete) then
	  psi1=GetPsi1(gvec,rho*z0+eps,k1)
      else
	  psi1=GetPsi2(gvec,rho*z0+eps,k1)
	endif

      if (IsNan(psi1)) then
	   PhiEps=psi1
	   Return
	Endif

      c1=psi1**(-1/eta)
	if (c1 .lt. 0.0) then
	 FError=1
	 ErrorType=4
	 PhiEps=0./0.
	 Return
	Endif

      k2=dexp(rho*z0+eps)*(k1**alpha)+(1-delta)*k1-c1

      if (k2 .lt. (1-delta)*k1) then
         c1=dexp(rho*z0+eps)*(k1**alpha)
         m1=c1**(-eta)-psi1
         if (m1 .lt. 0.0) then
	     FError=1
	     ErrorType=5
		 PhiEps=0./0.
		 Return
	   Endif
      else
         m1=0
      endif
      x=(c1**(-eta))*(1-delta + alpha*dexp(rho*z0+eps)*(k1**(alpha-1)))
     +  -(1-delta)*m1
      p=((2*pi*(sigma**2))**(-0.5))*(dexp((-0.5/(sigma**2))*(eps**2)))

      PhiEps=p*x

	Return

	End Function



C DMStat

      Subroutine DMStat(dm,chi)
                       
	use Def1	
	use IMSLF90
	
	Real(8) DM(1,1), Chi(1:2)

	Integer(4) t, t1, j, m
	Real(8) evec(1:nobs,1), z(1:nobs,1+2*lags),
     +        Amat(1:2*lags+1,1:2*lags+1),
     +        Bvec(1:2*lags+1,1), Avec(1:2*lags+1,1),
     +        Help(1:2*lags+1,1), DCHIIN
            
	m=1+2*lags
      t1=nobs+ta+lags-1

      Do t=ta+lags,t1
         
	   evec(t-ta-lags+1,1)=beta*(ct(t+1)**(-eta))*
     +	   (1-delta+alpha*dexp(zt(t+1))*
     +       (kt(t+1)**(alpha-1.)))- beta*mt(t+1)*(1-delta)-
     +       (ct(t)**(-eta)) + mt(t)
         z(t-ta-lags+1,1)=1.0
	   Do j=1,lags
	      z(t-ta-lags+1,1+j)=ct(t-j)
	      z(t-ta-lags+1,1+lags+j)=zt(t-j)
	   End Do

	End Do

      Avec=MatMul(Transpose(z),evec)
      Amat=0
	Do t=1,nobs
	   Help(1:m,1)=z(t,1:m)
	   Amat(1:m,1:m)=Amat(1:m,1:m)+(evec(t,1)**2)*
     +   MatMul(Help,Transpose(Help))
	End Do

	Call DLINRG(m,Amat,m,Amat,m)

      Bvec=MatMul(Amat,Avec)
	DM=MatMul(Transpose(Avec),Bvec)
     	
	chi(1)=DCHIIN(chi(1),dble(2*lags+1))
	chi(2)=DCHIIN(chi(2),dble(2*lags+1))

	Return

	End Subroutine      

C Simulate:

      Subroutine Simulate(Gamma,freq,hits,stat)
	                         
      use Def1

	Integer(4) freq(3), hits(2,2)
      Real(8) Gamma(ng),stat(3)


      Integer(4) t, i, j, t1, zindex(nobs+ta+lags+1)
	Real(8) rhs, dm, chi(2), eps(nobs+ta+lags+1), GetPsi1, GetPsi2,
     +        StDev

	External GetPsi1, GetPsi2      

	freq=0
	hits=0
	stat=0.0
	t1=nobs+ta+lags+1

C Loop over number of simulations
      Do i=1, nofs,1

C Draw shocks
         if (mca) then	      
	      Call MarkovTP(t1,nz,pmat,floor(dble(nz/2)),zindex)
	      Do j=1,t1
	        zt(j)=zgrid(zindex(j))
	      End Do
         else
	      Call DRNNOR(t1,eps) ! Draw shocks from a standard normal
	      zt(1)=0.0
            Do t=1,t1-1
	        zt(t+1)=zt(t)*Rho + Sigma*eps(t)
	      End Do
	      If (MinVal(zt) .lt. zmin) hits(1,1)=hits(1,1)+1
	      If (MaxVal(zt) .gt. zmax) hits(1,2)=hits(1,2)+1
	   Endif 
         
C Generate Time Series

      kt(1)=ks
	
      do t=1,t1

	    If (complete) then
		    rhs=GetPsi1(Gamma,zt(t),kt(t))
	    else
	        rhs=GetPsi2(Gamma,zt(t),kt(t))
          endif
	    if (IsNaN(rhs)) goto 2   
 	         ct(t)=rhs**(-1.0/eta)
           kt(t+1)=dexp(zt(t))*(kt(t)**alpha)+(1-delta)*kt(t)-ct(t)

C here comes the code that implements the Kuhn-Tucker conditions

        if ( kt(t+1) .lt. (1-delta)*kt(t) ) then
	          ct(t)=dexp(zt(t))*(kt(t)**alpha)
	          mt(t)=(ct(t)**(-eta))-rhs
	        kt(t+1)=(1-delta)*kt(t)
	  else
	      mt(t)=0.0
	  endif

	end do	

	if (MinVal(kt) .lt. kmin) hits(2,1)=hits(2,1)+1
	if (MaxVal(kt) .gt. kmax) hits(2,2)=hits(2,2)+1

      stat(1)=stat(1)+StDev(nobs,ct(ta+lags:t1-2))
	stat(2)=stat(2)+StDev(nobs,kt(ta+lags:t1-2))
	stat(3)=stat(3)+StDev(nobs,zt(ta+lags:t1-2))

      freq(3)=freq(3)+1

      chi(1)=0.025
	chi(2)=0.975

      Call DMStat(dm,chi)

      if (dm .lt. chi(1)) freq(1)=freq(1)+1
	if (dm .gt. chi(2)) freq(2)=freq(2)+1

2     Continue

      End Do
      stat(1)=stat(1)/dble(nofs)
	stat(2)=stat(2)/dble(nofs)
	stat(3)=stat(3)/dble(nofs)

	Return

	End Subroutine
      


C ARMC1: Computes the transition probability from state x0 to state x[i].

      Subroutine ARMC1(x0,p)

      use Def1
	use IMSLF90       

      Real(8), Intent(IN)   :: x0
	Real(8), Intent(OUT) :: p(1:nz)

      Integer(4) i
      Real(8) sy, w, dnordf, test(1:nz)

      sy=sigma/dsqrt(1.-(rho**2))
                  
      Call SeqaD(-size*sy,dble((2.*size*sy)/(nz-1)),nz,test) !liefert nur zgrid
	w=zgrid(2)-zgrid(1)
      
	p(1)=dnordf((zgrid(1)-rho*x0 + 0.5*w)/sigma)

      Do i=2,nz-1,1
	    
	   p(i)= dnordf((zgrid(i)-rho*x0 + 0.5*w)/sigma)
     +        -dnordf((zgrid(i)-rho*x0 - 0.5*w)/sigma)
		
	End Do

	p(nz)=1.0-Sum(p(1:nz-1))

      Return

      End Subroutine


