C The file holds a number of simple tools
C that we use in our main programs
C
C     - ElapsedTime
C     - GetD
C     - GetI
C     - GetI2
C     - GetC
C     - GetL
C     - MyDate
C     - StDev
C     - MyStat
C     - ChebEval2
C     - ChebEval3
C     - GC_Int1
C     - GC_Int2
C     - SeqaD
C     - MarkovAR
C     - MarkovTR
C****************************************************************
C
C ElapsedTime
C
C Purpose: returns an elapsed time string
C
C Usage: Call ElapsedTime(Seconds,Time)
C
C Input:   Seconds: Real(4), gives the
C          elapsed time in seconds
C
C
C Output:  Time: Character(100), holds the
C          formated time string
C          
      Subroutine ElapsedTime(Seconds,Time)

      Real(4) Seconds, totdays, tothrs, totmins, secs
	Integer days, hrs, mins
	Character(256) Time

      if (Seconds .lt. 0.0) then
	   Write(Time,"('Negative Time')")
	   Return
	endif	
	totdays = Seconds/(3600*24)
      days = floor(totdays)
	
      tothrs = (totdays-Real(days))*24
      hrs = floor(tothrs)
	
      totmins = (tothrs-real(hrs))*60
      mins = floor(totmins)
	
      secs = (totmins-real(mins))*60;
      
      if (days .gt. 1) then
        Write(Time,"(I1,' Days')") days
      endif

	if (days .eq. 1) then    
        Write(Time,"(I1,' Day')") days
      endif

      if (days .eq. 0) then
        Time=""
      endif    
      
	if (hrs .eq. 1) then
	  write(Time,"(A,I2,A)") trim(Time), hrs, ' hour '
      endif

	if (hrs .gt. 1) then
	  write(Time,"(A,I3,A)") trim(Time), hrs, ' hours '
      endif
	
	if (mins .eq. 1) then
	  write(Time,"(A,I2,A)") trim(Time), mins, ' minute '
      endif

      if (mins .gt. 1) then
	  write(Time,"(A,I3,A)") trim(Time), mins, ' minutes '
      endif
      
	write(Time,"(A,F6.2,A)") trim(Time), secs, ' seconds'            

      RETURN
	END Subroutine

C******************************************************************
C
C GetD
C
C Purpose: read a double precion variable from a text file
C
C Usage: Call GetD(VarName,FileName,Var)
C
C GetD(VarName,FileName,Var)
C
C Input: VarName:  Character*(*), the name of the variable
C                  in the file
C        FileName: Character*(*), the name of the file
C                  (must be in the current working directory
C
C Output: Var: the name of the variable that is asigned the
C              number read from the file
C
C Example: Suppose in the ASCII-File Parameters.txt
C          contains the following lines\
C          alpha=0.35
C          beta=0.99
C          delta=0.01
C          The command CALL GetD("delta","Parameters.txt",delta)
C          returns 0.01 and stores that value in the variable
C          delta.
C

      Subroutine GetD(String1,String2,Pvalue)
 
      use DFLIB
      Character*(*), INTENT(IN) :: String1, String2
	
      Real(8) Pvalue
      Character(256) Text, Rest, Suche
      Integer I1, I2
      	
      Suche=trim(String1)//"="
      I1=0
	I2=len_trim(Suche)

      open(unit=99,file=String2,err=104)
      
      
      Do while ((.not. eof(99)) .and. (I1 .eq. 0))
	    Read(99,'(A)') Text	    
	Rest=Text(I2+1:len(Text))
	Text=Text(1:I2)	
	If (trim(Suche) .eq. Text) then
	   I1=1
	   read(Rest,*) Pvalue
      End If
      end do
      if (I1 .eq. 0) then
	continue
	Pvalue=-1
      Text="Variable " // String1 // " not found in file " 
     +        //  String2 // "!"
	I2=MessageBoxQQ(Text,'Error',MB$SYSTEMMODAL )      
      endif
      close(99)
104   If (I1 .eq. 0) then
       Text="File not found"
	 I2=MessageBoxQQ(Text,'Error',MB$SYSTEMMODAL )
	End If
      Return      
	End Subroutine
	
C**********************************************************************
C
C GetI
C
C Purpose: Same as GetD, but returns an Integer value in
C Var
C
C Usage: Call GetI(VarName,FileName,Var)

      Subroutine GetI(String1,String2,Pvalue)

      use DFLIB 
      Character*(*), INTENT(IN) :: String1, String2
	
      Integer Pvalue
      Character(256) Text, Rest, Suche
      Integer I1, I2
      	
      Suche=trim(String1)//"="
      I1=0
	I2=len_trim(Suche)

      open(unit=99,action="Read",file=String2,err=105)
          
      Do while ((.not. eof(99)) .and. (I1 .eq. 0))
	    Read(99,'(A)') Text	    
	Rest=Text(I2+1:len(Text))
	Text=Text(1:I2)	
	If (trim(Suche) .eq. Text) then
	   I1=1
	   read(Rest,*) Pvalue
      End If
      end do
      if (I1 .eq. 0) then
	continue
	Pvalue=-1
      Text="Variable " // String1 // " not found in File " 
     +        //  String2 // " !"
      I2=MessageBoxQQ(Text,'Error',MB$SYSTEMMODAL)       
      endif
      close(99)
105   If (I1 .eq. 0) then
      Text="File not found"
      I2=MessageBoxQQ(Text,'Error',MB$SYSTEMMODAL)
	End If
      Return
	End Subroutine

C**********************************************************************
C
C GetI2
C
C Purpose: Same as GetD, but returns an Integer(2) value in
C Var
C
C Usage: Call GetI(VarName,FileName,Var)

      Subroutine GetI2(String1,String2,Pvalue)

      use DFLIB 
      Character*(*), INTENT(IN) :: String1, String2
	
      Integer(2) Pvalue
      Character(256) Text, Rest, Suche
      Integer I1, I2
      	
      Suche=trim(String1)//"="
      I1=0
	I2=len_trim(Suche)

      open(unit=99,action="Read",file=String2,err=108)
          
      Do while ((.not. eof(99)) .and. (I1 .eq. 0))
	    Read(99,'(A)') Text	    
	Rest=Text(I2+1:len(Text))
	Text=Text(1:I2)	
	If (trim(Suche) .eq. Text) then
	   I1=1
	   read(Rest,*) Pvalue
      End If
      end do
      if (I1 .eq. 0) then
	continue
	Pvalue=-1
      Text="Variable " // String1 // " not found in File " 
     +        //  String2 // " !"
      I2=MessageBoxQQ(Text,'Error',MB$SYSTEMMODAL)       
      endif
      close(99)
108   If (I1 .eq. 0) then
      Text="File not found"
      I2=MessageBoxQQ(Text,'Error',MB$SYSTEMMODAL)
	End If
      Return
	End Subroutine


C********************************************************************
C GetC
C
C Purpose: read a Character variable from a text file
C
C Usage: Call GetC(VarName,FileName,Var)
C
C
C Input: VarName:  Character*(*), the name of the variable
C                  in the file
C        FileName: Character*(*), the name of the file
C                  (must be in the current working directory
C
C Output: Var: the name of the variable that is asigned the
C              string read from the file
C
C Example: Suppose in the ASCII-File Parameters.txt
C          contains the following lines\
C          alpha=0.35
C          beta=0.99
C          file=outfile
C          The command CALL GetC("file","Parameters.txt",file)
C          returns the string 'outfile' in the variable file
C

      Subroutine GetC(String1,String2,String3)
 
      use DFLIB
      Character*(*), INTENT(IN) :: String1, String2
	Character*(*), INTENT(OUT) :: String3
      
      Character(256) Text, Rest, Suche
      Integer I1, I2
      	
      Suche=trim(String1)//"="
      I1=0
	I2=len_trim(Suche)

      open(unit=99,file=String2,err=106)
      
      
      Do while ((.not. eof(99)) .and. (I1 .eq. 0))
	    Read(99,'(A)') Text	    
	Rest=Text(I2+1:len(Text))
	Text=Text(1:I2)	
	If (trim(Suche) .eq. Text) then
	   I1=1
	   read(Rest,*) String3
      End If
      end do
      if (I1 .eq. 0) then
	continue
	String3=""
      Text="Variable " // String1 // " not found in file " 
     +        //  String2 // "!"
      I2=MessageBoxQQ(Text,'Error',MB$SYSTEMMODAL)	
      
      endif
      close(99)
106   If (I1 .eq. 0) then
        Text="File not found"
	  I2=MessageBoxQQ(Text,'Error',MB$SYSTEMMODAL)
	End If
      Return      
	End Subroutine
	

C**********************************************************************
C
C GetL
C
C Purpose: Same as GetD, but returns a logical value in
C Var
C
C Usage: Call GetL(VarName,FileName,Var)

      Subroutine GetL(String1,String2,Pvalue)
 
      use DFLIB
      Character*(*), INTENT(IN) :: String1, String2
	
      Logical Pvalue
      Character(256) Text, Rest, Suche
      Integer I1, I2
      	
      Suche=trim(String1)//"="
      I1=0
	I2=len_trim(Suche)

      open(unit=99,action="Read",file=String2,err=107)
          
      Do while ((.not. eof(99)) .and. (I1 .eq. 0))
	    Read(99,'(A)') Text	    
	Rest=Text(I2+1:len(Text))
	Text=Text(1:I2)	
	If (trim(Suche) .eq. Text) then
	   I1=1
	   read(Rest,*) Pvalue
      End If
      end do
      if (I1 .eq. 0) then
	continue
	Pvalue=-1
      Text="Variable " // String1 // " not found in File " 
     +        //  String2 // " !"
      I2=MessageBoxQQ(Text,'Error',MB$SYSTEMMODAL)
      endif
      close(99)
107   If (I1 .eq. 0) then
       Text="File not found"
	 I2=MessageBoxQQ(Text,'Error',MB$SYSTEMMODAL)
	End If
      Return
	End Subroutine


C***********************************************************************
C MyDate
C
C Purpose: Retrieve the current date and time and
C          returns it in a Character(50) string
C
C Usage: X=MyDate()

      CHARACTER(50) Function  MyDate()

	INTEGER(2) year,month,day, hour, min, sec, hsec
	CALL GETDAT(year,month,day)
	CALL GETTIM (hour, min, sec, hsec) 
	WRITE(MyDate,"(I2,'.',I2,'.',I4,'  ',I2,':',I2,':',
     +      I2)") day,month,year,hour,min,sec

      Return
	END FUNCTION


C*********************************************************************************
C StDev
C
C Computes the standard deviation of the elemtens in the row vector x(n)
C
C Usage: Result=StDev(n,x)
C
C        Input: n, Integer(4), the number of elements in x
C               x, Real(8), n times 1 vector
C
C        Output: Result, Real(8)


      Real(8) Function StDev(n,x)

	INTEGER(4) n
	REAL(8)  x(1:n)

      Integer(4) i
      Real(8) xmean, xsum  
      
	xsum=SUM(x)
	xmean=dble(Sum(x)/n)
	StDev=0.0
	Do i=1,n,1
        StDev=StDev+((x(i)-xmean)**2)
      End Do
	StDev=DSQRT(StDev/(n-1))
      Return
	END Function

C ***************************************************************************************
C MyStat
C
C Computes Minimum, Maximum and Standard Deviation of the
C Elements of a Vector
C
C Useage: Call MyStat(n,x,stat)
C
C Input: N: Integer(4), number of elements in x
C        X: Real(8), N times 1 vector
C
C Output: stat, Real(8), 4 times 1 vector, where
C              stat(1)=minimum of x
C              stat(2)=maximum of x
C              stat(3)=mean of x 
C              stat(4)=standard deviation of x


      Subroutine MyStat(n,x,stat)

	INTEGER(4) n, i
	REAL(8)  x(1:n), stat(4)
       
      stat(1)=MinVal(x)
	stat(2)=MaxVal(x)      
	stat(3)=dble(Sum(x)/n)
	stat(4)=0.0
	Do I=1,n
        stat(4)=stat(4)+(x(i)-stat(3))**2
      End Do
	stat(4)=DSQRT(dble(stat(4)/(n-1)))
      Return
	END SUBROUTINE


C **************************************************************************************************
C ChebEval2
C
C Evaluates a second degree complete polynomial
C
C Usage: y=ChebEval2(gvec,ng,z,bounds)
C
C Input: gvec  := Real(8)    ng times 1 vector, coefficients of the polynomial
C        degree:= Integer(4) degree of the polynomial
C        z     := Real(8)    2 times 1 vector, point, where the polynomial is to be evaluated
C        bounds:= Real(8)    2 times 2 matrix, first row: lower/upper bound for z1,
C                                             second row: lower/upper bound for for z2
C
C Output:  y   := Real(8)    value of the polynomial at point x
C

      Real(8) Function ChebEval2(gvec,degree,z,bounds)

      Integer(4) degree
	Real(8)  gvec((degree+1)*(degree+2)/2), z(1:2), bounds(1:2,1:2)

      Integer(2) p, i, j, k
      Real(8) t1(1:degree+1), t2(1:degree+1), y, x(1:2)
      
      p=degree+1
      x(1)=((2.0*(z(1)-bounds(1,1)))/(bounds(1,2)-bounds(1,1))) - 1.0 ! Map z1 into [-1,1] 
      x(2)=((2.0*(z(2)-bounds(2,1)))/(bounds(2,2)-bounds(2,1))) - 1.0 ! Map z2 into [-1,1] 

      t1(1)=1.0
	t2(1)=1.0
	t1(2)=x(1)
	t2(2)=x(2)

      Do i=3,p,1
  
       t1(i)=2.0*x(1)*t1(i-1)-t1(i-2)
       t2(i)=2.0*x(2)*t2(i-1)-t2(i-2)
         
      End do

      y=0.0
	k=1
      Do i=1,p,1
         Do j=1,(p+1-i),1
            y=y+gvec(k)*t1(j)*t2(i)
	      k=k+1
         End Do
      End Do
      ChebEval2=y

	Return
      End Function


C **************************************************************************************************
C ChebEval3
C
C Evaluates a second degree product base polynomial
C
C Usage: y=ChebEval3(gvec,degree,z,bounds)
C
C Input: gvec  := Real(8)    ng times 1 vector, coefficients of the polynomial
C        degree:= Integer(4) 2 times  1 vector, degree(1): degree of polynomial in z1
C                                               degree(2): degree of polynomial in z2
C        z     := Real(8)    2 times 1 vector, point, where the polynomial is to be evaluated
C        bounds:= Real(8)    2 times 2 matrix, first row: lower/upper bound for z1,
C                                             second row: lower/upper bound for for z2
C
C Output:  y   := Real(8)    value of the polynomial at point x
C

      Real(8) Function ChebEval3(gvec,degree,z,bounds)

      Integer(4) degree(2)
	Real(8)  gvec(degree(1),degree(2)), z(1:2), bounds(1:2,1:2)

      Integer(2) i, j
      Real(8) t1(1:degree(1)), t2(1:degree(2)), y, x(1:2)
      
      
      x(1)=((2*(z(1)-bounds(1,1)))/(bounds(1,2)-bounds(1,1))) - 1. ! Map z1 into [-1,1] 
      x(2)=((2*(z(2)-bounds(2,1)))/(bounds(2,2)-bounds(2,1))) - 1. ! Map z2 into [-1,1] 

      t1(1)=1.
	t2(1)=1.
	t1(2)=x(1)
	t2(2)=x(2)

      Do i=3,degree(1),1
  
       t1(i)=2*x(1)*t1(i-1)-t1(i-2)       
         
      End do

	Do i=3,degree(2),1

	  t2(i)=2*x(2)*t2(i-1)-t2(i-2)

	End Do

      y=0

      Do i=1,degree(1),1
         Do j=1,degree(2),1
            y=y+gvec(i,j)*t1(i)*t2(j)	      
         End Do
      End Do
      ChebEval3=y

	Return
      End Function


C *****************************************************************************************************
C GC_Int1
C
C Integrate a real valued function over [a,b], using Gauss-Chebyshev quadrature
C
C Usage:  y=GC_Int(f, d, n)
C
C Input: f := Real(8) pointer to the procedure that evaluates the function, usage must be y=f(x)
C        d := Real(8) 2 times 1 vector, d(1) lower, d(2) upper limit of integration
C        n := Integer(2), the number of nodes where f is to be evaluated in order to compute y
C
C Ouput: y := Real(8), the value of the integral of f(x) over [a,b]


      Real(8) Function GC_Int1(f,d,n)

	use IMSLF90      

      Integer(4) n
      Real(8) f, d(2)

      Integer(4) i
      Real(8) x(n), z(n), sum, pi, dconst
		
      External f, dconst

      pi=Dconst('pi')

      Do i=1,n

         x(i)=dcos(((2.*i-1)/(2.*n))*pi)         ! zero of n-the degree Chebyshev polynomial
	   z(i)=(x(i)+1.)*(d(2)-d(1))*0.5 + d(1)  ! adjusted to [a,b]
      
	End Do
     
      sum=0
     
      do i=1,n
         sum=sum+f(z(i))*dsqrt(1.-(x(i)**2))          
      End do
   
      sum=pi*(d(2)-d(1))*sum
      sum=sum/(2.*n)
    
      GC_Int1=sum
      
	Return

	End Function


C *************************************************************************************************************
C GC_Int2
C
C Integrate a real valued function f(x,y) over [a,b]x[c,d], using Gauss-Chebyshev quadrature
C
C Usage:  y=GC_Int(f, d, n)
C
C Input: f := Real(8) pointer to the procedure that evaluates the function, usage must be z=f(x,y)
C        d := Real(8) 2 times 2 matrix, d(1,1) lower, d(1,2) upper limit of integration for x,
C                                       d(2,1) lower, d(2,2) upper limit of integration for y
C        n := Integer(4) 2 times 1 vector, n[1] the number of nodes for x, n[2] the number of node for y
C
C   Output:  Real(8) v := scalar: the (approximate) value of the integral
C
C   Remarks: if f(x,y) returns a scalar missing value (since f is not computable at this point)
C            the procedure stops and returns 


      Real(8) Function GC_Int2(f,d,n)

	use IMSLF90      

	Integer(4) n(2)
	Real(8) f, d(2,2)

      Integer(4) i, j
	Real(8)  x1(n(1)), x2(n(2)), z1(n(1)), z2(n(2)), y, sum, dconst, pi

      External f, dconst

      pi=dconst('pi')
      Do i=1,n(1),1

         x1(i)=dcos(((2.*i - 1.)/(2.*n(1)))*pi)             ! zero of the n-th degree Chebyshev polynomila for x
         z1(i)=(x1(i)+1.)*(d(1,2)-d(1,1))*0.5 + d(1,1) ! adjusted to [a,b]                             
   
      End Do

      Do i=1,n(2)
         x2(i)=dcos(((2.*i-1.)/(2.*n(2)))*pi)             ! zeros of the n-th degree Chebyshev polynomial for y 
         z2(i)=(x2(i)+1.)*(d(2,2)-d(2,1))*0.5 + d(2,1) ! adjusted to [c,d]                             

      End Do
      open(20,file='Test.txt')
      sum=0
      Do i=1,n(1),1
        Do j=1,n(2),1        
            y=f(z1(i),z2(j))
	      write(20,"(3F15.10)") z1(i), z2(j), y
            if (isnan(y)) then
		     GC_INT2=y
			 Return
		  Endif
            sum=sum+y*dsqrt(1.-(x1(i)**2))*dsqrt(1.-(x2(j)**2))              
        End Do        
      End Do
      close(20)
      sum=pi*(d(1,2)-d(1,1))*pi*(d(2,2)-d(2,1))*sum
      sum=sum/(2.*n(1)*2.*n(2))
   
      GC_INT2=sum

	Return

	End Function

C*********************************************************************
C SeqaD
C
C Puropse: creates an additive series of double precion
C
C Usage Call SeqaD(xmin, step, nobs,x)
C
C Inputs:
C 
C xmin: Real(8), the first element in the series
C step: Real(8), increment
C nobs: Integer(4), number of elements in the series
C
C Output:
C
C x: Real(8) vector with nobs elements, where
C   x(i)=xmin+(i-1)*step


      Subroutine SeqaD(xmin,step,nobs,x)

	Integer(4) nobs
      Real(8)    xmin, step, x(nobs)
	      
      Integer(4) i

	x(1)=xmin
	
	do i=2,nobs
	 x(i)=x(i-1)+step	
      end do     
      
	Return
	End Subroutine

C**********************************************************************
C
C MarkovAR
C
C Purpose: Approximate AR-Process by Markov Chain
C
C Usage: Call MarkovAR(size,m,rho,sigma,z,p)
C
C Input: size, scalar, Real(8), fraction of
C                      unconditional variance of AR process,
C                      determines grid size
C
C        m, Integer(4), number of states of Markov Chain
C
C      rho, Real(8), autoregressiv parameter
C  
C    sigma, Real(8), standard deviation of innovations
C
C Output: z, m x 1, Real(8) vector, states of the Markov Chain
C
C         p, m x m, Real(8) transition matrix
C
      Subroutine MarkovAR(size,m,rho,sigma,z,p)

      use IMSLF90       

	Integer(4) m
	Real(8) size, rho, sigma, z(1:m), p(1:m,1:m)

      Real(8) sigmaz, zbar, dnordf
	Integer(4) I, J
		 
      sigmaz=dsqrt(sigma**2/(1-(rho**2)))
      zbar=size*sigmaz

      Call SeqaD(-zbar,dble(2.0*zbar/(m-1)), m,z)
  
      Do I=1,m
	
         p(i,1)=dnordf((z(1)-rho*z(i)+(z(2)-z(1))/2)/sigma)
         
	   Do j=2,m-1
   
           p(i,j)=dnordf((z(j)-rho*z(i)+(z(j)-z(j-1))/2)/sigma)-
     +             dnordf((z(j)-rho*z(i)-(z(j)-z(j-1))/2)/sigma)
           continue
	   End Do
	   p(i,m)=1-SUM(p(i,1:m-1))
	   
	End Do  
   
      End Subroutine



C*********************************************************************
C
C MarkovTP
C
C Purpose: Compute realization of Markov chain
C
C Usage: Call MarkovTP(n,m,p,i,zt)
C
C Input: n, Integer(4), length of time path
C    
C        m, Integer(4), size of transition matrix
C
C        p, Real(8), m x m Transition matrix
C
C        i0, Integer(2), index of first realization
C
C Ouptut: zt, Integer(4), n x 1 vector of indices from
C             the set 1, 2, 3, ..., m
C

      Subroutine MarkovTP(n,m,p,i0,zt)

      use IMSLF90       

	Integer(4) n, m, i0, zt(1:n)
	Real(8) p(1:m,1:m)

      Integer(4) i, j, l
      Real(8) eps(1:n), pu, po


C Draw a sequence of uniformly distributed random numbers
      Call DRNUN(n,eps)
 
      zt(1)=i0
	
	Do 10, I=1,N-1	
       j=zt(i)	 
       if (eps(i) .le. p(j,1)) then
             zt(i+1)=1	
             goto 10
       endif		   	       
       pu=p(j,1)
	 po=pu
	 Do 20, l=2,m
                 pu=po
                 po=pu+p(j,l)	
                 if ((eps(i) .gt. pu) .and. (eps(i) .le.po)) then
		         zt(i+1)=l
		         goto 20
			   endif
20    continue      

10    Continue

      End Subroutine