C Version of the program with q and E(c1-bc) approximated by Chebyshev polynomials
C
C
C
C Epp.for (Version Epp_4.for is the most recent one (24.03.04)

C Fortran program to solve the equity premium puzzle model from Herr/Maussner, Chapter 4

C   Module definitions:

C Parameters of the model and of the algorithm
      Module Def1
	
	Real(8) alpha, beta, b, eta, delta, rho, sigma, zeta, a1, a2
	Real(8) zmin, zmax, kmin, kmax, cmin, cmax, ks, cs, zs,ls
	Real(8) bounds1(1:3,1:2)
      Real(8), allocatable :: gvec(:,:), zgrid(:), pmat(:,:)
      Real(8), allocatable :: ct(:), kt(:), zt(:), qt(:), rt(:,:), lt(:)

      Integer(4) degree, npar

C Parameters of the algorithm     	
	Integer(4) nz, nobs, nofs, ta, lags
	Integer(4) znodes, knodes, cnodes
	Real(8)  size

	End Module Def1

C Module for possible Errors that need to be known by various subroutines
      Module Errors

	Integer(2) FError, ErrorType

	End Module Errors


C Here begins the main program
      program Epp

      use DFLIB
	use IMSLF90	
      use Def1
	use Errors
	use MNR
	      
      Integer(4) i, j, tmax, GetNPar, Search, hit(1:2,1:2), method
	Real(4) Asec
      Real(8), Allocatable :: Gamma1(:), Gamma2(:), Gamma3(:), fx(:)
	Real(8) StDev, fit, crit(1:6), xs(1:3), GetPsi, rs(1:3)

	Character(256) Laufzeit
	Character(50) MyDate	

	Logical initialize, increase, ok, InitialSettings

      
	External f1, f2, sys1, sys0, sys2

	Type (rccoord) curpos		

      
C Read Parameters from input file and compute derived parameters
C Parameters of the algorithm
      Call GetL('initialize','Parameters.txt',initialize)      
	Call GetL('increase','Parameters.txt',increase)
	Call GetI('search','Parameters.txt',search)
	Call GetI('degree','Parameters.txt',degree)	
      Call GetI('nobs','Parameters.txt',nobs)
	Call GetI('nofs','Parameters.txt',nofs)
	Call GetI('ta','Parameters.txt',ta)
	Call GetI('lags','Parameters.txt',lags)
      Call GetD('size','Parameters.txt',size)
	Call GetI('nz','Parameters.txt',nz)
	Call GetI('method','Parameters.txt',method)
      Call GetD('kmin','Parameters.txt',kmin)
	Call GetD('kmax','Parameters.txt',kmax)
	Call GetD('cmin','Parameters.txt',cmin)
	Call GetD('cmax','Parameters.txt',cmax)
      
C Parameters of the model
	Call GetD('alpha' ,'Parameters.txt',alpha)
	Call GetD('beta'  ,'Parameters.txt',beta )
	Call GetD('b'     ,'Parameters.txt',b)
	Call GetD('eta'  ,'Parameters.txt',eta ) 
	Call GetD('delta','Parameters.txt',delta)	
	Call GetD('zeta', 'Parameters.txt',zeta)
	zeta=1./zeta
	Call GetD('rho'  ,'Parameters.txt',rho  )
	Call GetD('sigma','Parameters.txt',sigma)

	Call GetI('znodes','Parameters.txt',znodes)
	Call GetI('knodes','Parameters.txt',knodes)
	Call GetI('cnodes','Parameters.txt',cnodes)
      
C Allocate space
      if (increase) then
	  i=GetNpar(degree-1)
	else
	  i=GetNpar(degree)
	endif
	Allocate(Gamma3(1:2*i))
	
      npar=GetNpar(degree)

C      tmax=nobs+ta+1
      tmax=nobs

      Allocate (gvec(1:npar,1:2),gamma1(1:2*npar),
     +  gamma2(1:2*npar), lt(1:tmax+1), rt(1:tmax,1:3),
     +  ct(1:tmax+1), zt(1:tmax+1), kt(1:tmax+2),
     + qt(1:tmax+1), zgrid(1:nz),pmat(1:nz,1:nz),fx(1:2*npar))

C Prepare output window
      ok=InitialSettings()
      Call MakeSmallWindow(10,"Program Progress",2)

C open logfile
      Asec=Secnds(0.0) ! get initial time

      open (11,file='LogFile.txt')
	WRITE(11,"(A50)") MyDate()		

C Compute bounds
      ks=((alpha*beta)/(1-beta*(1-delta)))**(1/(1-alpha)) ! deterministic steady state for the stock of capital     
      cs=(ks**alpha) - delta*ks                           ! deterministic steady state for the level of consumption 
      zs=0.0                                              ! unconditional mean of z 
	ls=(1-beta*b)*(((1-b)*cs)**(-eta))

	Call MarkovAR(size,nz,rho,sigma,zgrid,pmat)

      kmin=kmin*ks
      kmax=kmax*ks
      cmin=cmin*cs
      cmax=cmax*cs
	zmin=zgrid(1)
	zmax=zgrid(nz)

	bounds1(1,1)=zmin
	bounds1(1,2)=zmax
	bounds1(2,1)=dlog(kmin)
	bounds1(2,2)=dlog(kmax)
	bounds1(3,1)=dlog(cmin)
	bounds1(3,2)=dlog(cmax)

      a1=delta**zeta
	a2=-(zeta/(1.0-zeta))*delta

C Fix starting time
      Asec=Secnds(0.0)
777   Continue
C Initialize 
      if (.not. initialize) then
C        open(20,file='ParIni.txt',form='formatted')
	  open(20,file='ParIni.bin',form='unformatted')
       read(20) gamma3
C	  read(20,*) gamma3(1:2*Npar)
	  close(20)	  
	  if (increase) then	    
	    Call GetGamma(degree-1,i,npar,Gamma3(1:i),Gamma1(1:npar))
		Call GetGamma(degree-1,i,npar,Gamma3(1+i:2*i),
     +		Gamma1(1+npar:2*npar)) 
        else
	    Gamma1=Gamma3     
	   endif
      else
	  Select Case(method)
	    Case(1)
	      if (Search.eq.1) Call GSearch1(F1,2*npar,Gamma1,fit)
	      if (Search.eq.2) Call GSearch2(F1,2*npar,Gamma1,fit)
	    Case(2)
	      if (Search.eq.1) Call GSearch1(F2,2*npar,Gamma1,fit)
	      if (Search.eq.2) Call GSearch2(F2,2*npar,Gamma1,fit)
	  End Select
	endif

776   continue
      if (initialize) then
        Call ClearScreen($GCLEARSCREEN)        
	  CALL SETTEXTPOSITION(1,2,curpos)
        Write(10,"(A,2F15.10,2X)") 'Function value of best solution:',
     +   fit
	  Write(11,"(A)") 'Solution from search:'
	  Write(11,"(A,F15.10)") 'Function value of best solution:', fit
	  Call PrintSolution(gamma1)
      endif

C Refine solution
      Select Case(method)
	   Case(1)
          	Call FixvMN(2*npar,Gamma1,Sys1,Gamma2,crit)
	   Case(2)
              Call FixvMN(2*npar,Gamma1,Sys2,Gamma2,crit)           
      End Select

	if (crit(1) .eq. 0.0) then
	   Call ClearScreen($GCLEARSCREEN)        
    	   CALL SETTEXTPOSITION(1,2,curpos)
	   Write(10,"(A)") 'Normal Termination from FixvMN'
	   Write(11,"(A)") 'Normal Termination from FixvMN'
	endif
	
	if (crit(1) .ne. 0.0) then
	    Call ClearScreen($GCLEARSCREEN)        
    	    CALL SETTEXTPOSITION(1,2,curpos)
          Write(10,"(A)") MNR_Message(IDINT(crit(1)))
	    Write(11,"(A)") MNR_Message(IDINT(crit(1)))	
	    if (initialize) goto 777
	endif
	Write(10,"(A,F15.8)") 'Max(abs(fx))       = ', crit(2)
	Write(11,"(A,F15.8)") 'Max(abs(fx))       = ', crit(2)
	Write(10,"(A,F15.8)") 'Max(|dx/x|)        = ', crit(3)
	Write(11,"(A,F15.8)") 'Max(|dx/x|)        = ', crit(3)
	Write(10,"(A,F15.8)") 'Max. Relgrad       = ', crit(4)
	Write(11,"(A,F15.8)") 'Max. Relgrad       = ', crit(4)
	Write(10,"(A,F15.8)") '0.5*fx*fx          = ', crit(5)
	Write(11,"(A,F15.8)") '0.5*fx*fx          = ', crit(5)	    
	Write(10,"(A,I5)")  'Iterations         = ', IDINT(crit(6))
	Write(11,"(A,I5)")  'Iterations         = ', IDINT(crit(6))	   	   
      if (crit(1) .ne. 0.0) then
	   Call GetTextPosition(curpos)
	   Call SetTextPosition(curpos.row+1,2,curpos)
	   Write(10,"(A)") 'Save Solution (.T./.F.):'
	   Call SetTextPosition(curpos.row+1,24,curpos)
	   Read(10,"(A1)") ok
	   if (ok) then
	     open(12,file='ParIni.bin',form='unformatted')
	     write(12) Gamma2
	     close(12)
	   endif
	endif

	if (crit(1) .eq. 0.0) then
	  Write(11,"(A)") 'Solution:'
        Call PrintSolution(Gamma2)
	  open(12,file='ParIni.bin',form='unformatted')
	  write(12) Gamma2
	  close(12)

        xs(1)=beta*GetPsi(Gamma2(1:npar),zs,ks,cs)
	  xs(2)=GetPsi(Gamma2(1+npar:2*npar),zs,ks,cs)
	  xs(2)=(((1-b)*cs)**(-eta))-beta*b*xs(2)
	  Call Gettextposition(curpos)
        Call SetTextPosition(curpos.row+1,2,curpos)
	  Write(10,"(A,1X,F10.6)") 'Implied q= ',xs(1)
	  Write(11,"(A,1X,F10.6)") 'Implied q= ',xs(1)
	  Call GetTextPosition(curpos)
	  Call SetTextPosition(curpos.row,2,curpos)
	  Write(10,"(2(A,1X,F10.6))") 'Implied lambda= ',xs(2),
     +    'Stationary lambda= ', ls   
	  Write(11,"(2(A,1X,F10.6))") 'Implied lambda= ',xs(2),
     +    'Stationary lambda= ', ls
        i=MESSAGEBOXQQ ('Solve for ks and cs?','User Input',
     +    MB$SYSTEMMODAL .or. MB$YESNO) 
	  if (i.ne.MB$IDYES) then
	    goto 778	
	  endif 
        
C Compute stationary solution implied by gvec	  
	  Gamma1(1)=1
	  Gamma1(2)=ks
	  Gamma1(3)=cs
        i=3
	  gvec(1:npar,1)=Gamma2(1:npar)
	  gvec(1:npar,2)=Gamma2(1+npar:2*npar)
     	  Call FixvMN(i,Gamma1(1:3),Sys0,xs,crit)
 	  
	  Call GetTextPosition(curpos)
	  Call SetTextPosition(curpos.row,2,curpos)
	  Write(10,"(A,1X,F10.5)") 'qs1= ', xs(1)
	  Write(11,"(A,1X,F10.5)") 'qs1= ', xs(1)
	  Call GetTextPosition(curpos)
        Call SetTextPosition(curpos.row,2,curpos)
	  Write(10,"(2(A,2X,F10.5,2X))") 'ks1= ', xs(2),
     +     'ks= ', ks
	  Write(11,"(2(A,2X,F10.5,2X))") 'ks1= ', xs(2),
     +     'ks= ', ks
	  Call GetTextPosition(curpos)
	  Call SetTextPosition(curpos.row,2,curpos)
	  Write(10,"(2(A,2X,F10.5,2X))") 'cs1= ', xs(3),
     +     'cs= ', cs
	  Write(11,"(2(A,2X,F10.5,2X))") 'cs1= ', xs(3),
     +     'cs= ', cs
        Call Impulse(2*npar,Gamma2,25,xs,tmax)
	  Call Simulate(Gamma2,xs,hit,rs,fit)
	  Call GetTextPosition(curpos)
	  Call SetTextPosition(curpos.row,2,curpos)
	  Write(10,"(A,1X,F10.5)") 'Riskless rate=  ', rs(1)
        Write(11,"(A,1X,F10.5)") 'Riskless rate=  ', rs(1)
	  Call GetTextPosition(curpos)
	  Call SetTextPosition(curpos.row,2,curpos)
	  Write(10,"(A,1X,F10.5)") 'Equity rate=    ', rs(2)
	  Write(11,"(A,1X,F10.5)") 'Equity rate=    ', rs(2)
	  Call GetTextPosition(curpos)
	  Call SetTextPosition(curpos.row,2,curpos)	  	  
	  Write(10,"(A,1X,F10.5)") 'Equity premium= ', rs(3)
	  Write(11,"(A,1X,F10.5)") 'Equity premium= ', rs(3)	  
	  Call GetTextPosition(curpos)
	  Call SetTextPosition(curpos.row,2,curpos)
        Write(10,"(A,1X,F10.5)") 'Stationary rate=', ((1./beta)-1.)
	  Write(11,"(A,1X,F10.5)") 'Stationary rate=', ((1./beta)-1.)
	  Call GetTextPosition(curpos)
        Call SetTextPosition(curpos.row,2,curpos)
	  Write(10,"(A,2(I4,2X))") 'c<cmin or c>cmax:', hit(1,1), hit(1,2)
	  Call GetTextPosition(curpos)
	  Call SetTextPosition(curpos.row,2,curpos)
	  Write(10,"(A,2(I4,2X))") 'k<kmin or k>kmax:', hit(2,1), hit(2,2)
       Write(11,"(A,2X,F10.5)") 'Standard deviation of filtered output',
     + fit
        Write(11,"(A)") 'Standard deviation       Minium        Maximum'
        Write(11,"(A,2X,3(F10.5,2X))") 'Shock:       ',
     +      StDev(nobs+1,zt),MinVal(zt),MaxVal(zt)	  
        Write(11,"(A,2X,3(F10.5,2X))") 'Capital:      ',
     +      StDev(nobs+2,kt),MinVal(kt),MaxVal(kt)
        Write(11,"(A,2X,3(F10.5,2X))") 'Consumption:  ',
     +      StDev(nobs+1,ct),MinVal(ct),MaxVal(ct)
        Write(11,"(A,2X,3(F10.5,2X))") 'q:            ',
     +      StDev(nobs+1,qt),MinVal(qt),MaxVal(qt)


        Call Writeqck(Gamma2)

778     continue
	  Write(11,"(A)") 'Parameters:'
        Write(11,"('alpha=  ',F9.5)") alpha
	  Write(11,"('beta =  ',F9.5)") beta
	  Write(11,"('b    =  ',F9.5)") b
	  Write(11,"('eta  =  ',F9.5)") eta
	  Write(11,"('delta=  ',F9.5)") delta
	  Write(11,"('zeta =  ',F9.5)") (1./zeta)
	  Write(11,"('rho  =  ',F9.5)") rho
	  Write(11,"('sigma=  ',F9.5)") sigma

        Write(11,"('Parameters of the algorithm')")
	  Write(11,"('method= ',I1)") method
	  Write(11,"('degree= ',I1)") degree
	  if (method.eq.2) then
	     Write(11,"('znodes= ',I2)") znodes
		 Write(11,"('knodes= ',I2)") knodes
		 Write(11,"('cnodes= ',I2)") cnodes
	  Endif	 
	  Write(11,"('size     =  ',F7.3)") size
	  Write(11,"('kmin     =  ',F7.3)") (kmin/ks)
        Write(11,"('kmax     =  ',F7.3)") (kmax/ks)
	  Write(11,"('cmin     =  ',F7.3)") (cmin/cs)
	  Write(11,"('cmax     =  ',F7.3)") (cmax/cs)

	  Write(11,"('Parameters of the simulations:')")	  
	  Write(11,"('nobs=  ',I6)") nobs
	  Write(11,"('ta  =  ',I6)") ta
        Write(11,"('lags=  ',I6)") lags	   
	  Write(11,"('nofs=  ',I8)") nofs

	  
	           
	 
	endif

      close(11)
	end





C ************************************************ Subroutines ***************************************************

C f1: This function returns the maximum absolute value returned by Sys1. It is used
C     by the search routines.

      Real(8) Function f1(n,par)

      use Errors

	Integer(4) n
	Real(8) par(n)

	Real(8) fx(n)

      FError=0
	Call Sys1(n,n,par,fx)

      if (FError .ne. 0) then
	   F1=-1.0
	else
C         F1=MaxVal(dabs(fx))
        F1=dsqrt(Dot_Product(fx,fx))
	endif

      Return

	End Function

C f2: This function returns the maximum absolute value returned by Sys2. It is used
C     by the search routines.

      Real(8) Function f2(n,par)

      use Errors

	Integer(4) n
	Real(8) par(n)

	Real(8) fx(n)

      FError=0
	Call Sys2(n,n,par,fx)

      if (FError .ne. 0) then
	   F2=-1.0
	else
C         F1=MaxVal(dabs(fx))
        F2=dsqrt(Dot_Product(fx,fx))
	endif

      Return

	End Function



C Sys0: The zero of the rhs returned by this subroutine defines
C       the stationary solution of the approximate model solution
       
	Subroutine Sys0(n,m,x,fx)

	use Def1
      use Errors
      
	Integer(4) n, m
	Real(8) x(n), fx(m)

	Real(8) i, GetInv, GetPhi, GetPsi
      
      if (x(1).lt.0.0) FError=1
	if ((x(2).lt.kmin).or.(x(2).gt.kmax)) FError=2
	if ((x(3).lt.cmin).or.(x(3).gt.cmax)) FError=3
	if (FError.ne.0) then
	  fx=1
	  return
	Endif
	fx(1)=x(1)-beta*Getpsi(gvec(1:npar,1),zs,x(2),x(3))
       i=GetInv(x(1),x(2))
      fx(2)=x(2)-GetPhi(i,x(2))*x(2)-(1-delta)*x(2)
      fx(3)=x(3)-(dexp(zs)*(x(2)**alpha))+i
	
      Return

	End Subroutine

      
C Sys1: The zero of the rhs returned by this subroutine defines the
C       collocation solution of the model

      Subroutine Sys1(m,n,par,fx)

	use Def1
	use IMSLF90	
      use Errors

      Integer(4) m, n
	Real(8) par(n), fx(m), fx2(1:npar,1:2)

      Integer(4) i, i1, i2, i3, p1, k
	Real(8) zbar(degree+1),kbar(degree+1), cbar(degree+1), pi, Dconst
      Real(8) Resid(1:2)	

	pi=Dconst('pi')    
  
      p1=degree+1
	
      Do i=1,p1
	   zbar(i)=dcos(((dble(2.0*i)-1.0)/(dble(2.*p1))) *pi)
	   zbar(i)=0.5*(zbar(i)+1.0)*(zmax-zmin) + zmin
	End Do
	
	Do i=1,p1
	   kbar(i)=dcos(((dble(2.0*i)-1.0)/(dble(2.*p1)))*pi)   
	   kbar(i)=0.5*(kbar(i)+1.0)*(kmax-kmin) + kmin
	End Do

      Do i=1,p1
	   cbar(i)=dcos(((dble(2.0*i)-1.0)/(dble(2.*p1)))*pi)
	   cbar(i)=0.5*(cbar(i)+1.0)*(cmax-cmin) + cmin
	End Do      

	gvec(1:npar,1)=par(1:npar)
	gvec(1:npar,2)=par(1+npar:2*npar)      

      p1=degree+1
	
	k=1
      Do i1=1,p1
            Do i2=1,(p1+1-i1)
               Do i3=1,(p1+2-i1-i2)
                  Call GetResid(zbar(i1),kbar(i2),cbar(i3),Resid)
	            if FError.ne.0) Return
                  fx2(k,1:2)=Resid(1:2)
                  k=k+1      
               End Do
	      End Do
      End Do
      
      fx(1:npar)=fx2(1:npar,1)
	fx(1+npar:2*npar)=fx2(1:npar,2)

	Return

	End Subroutine

C Sys2: the zero of fx returned by this subroutine defines the Galerkin solution of
C       the model

      Subroutine Sys2(m,n,par,fx)

	use Def1
	use IMSLF90	
      use Errors

      Integer(4) m, n
	Real(8) par(n), fx(m), fx2(1:npar,1:2)

      Integer(4) i, i1, i2, i3, p1, k, j1, j2, j3
	Real(8) zbar(znodes,2), kbar(knodes,2), cbar(cnodes,2)
	Real(8) pi, Dconst, Resid(1:2), Cheb	

	pi=Dconst('pi')    
  
      p1=degree+1
	
      Do i=1,znodes
	   zbar(i,1)=dcos(((dble(2.0*i)-1.0)/(dble(2.*znodes))) *pi) ! zeros of 
	   zbar(i,2)=0.5*(zbar(i,1)+1.0)*(zmax-zmin) + zmin
	End Do
	
	Do i=1,knodes
	   kbar(i,1)=dcos(((dble(2.0*i)-1.0)/(dble(2.*knodes)))*pi)   
	   kbar(i,2)=0.5*(kbar(i,1)+1.0)*(kmax-kmin) + kmin
	End Do

      Do i=1,cnodes
	   cbar(i,1)=dcos(((dble(2.0*i)-1.0)/(dble(2.*cnodes)))*pi)
	   cbar(i,2)=0.5*(cbar(i,1)+1.0)*(cmax-cmin) + cmin
	End Do      

	gvec(1:npar,1)=par(1:npar)
	gvec(1:npar,2)=par(1+npar:2*npar)      

      p1=degree+1
	
	fx2=0.0
	open(21,file='Text.txt')
      Do i1=1,znodes
          Do i2=1,knodes
             Do i3=1,cnodes
                  Call GetResid(zbar(i1,2),kbar(i2,2),cbar(i3,2),Resid)
	            If (FError.ne.0) Return
C	Write(21,"('i1= ',I2,1X,'i2= ',I2,1X,'i3= ',I2,1X,5(F10.5,2X),1X,
C     + I1)")
C     + i1,i2,i3,zbar(i1,2),kbar(i2,2),cbar(i3,2),Resid(1),Resid(2),
C     + FError
	            k=1
	            Do j1=1,p1
	               Do j2=1,(p1+1-j1)
	                  Do j3=1,(p1+2-j1-j2)
				         fx2(k,1)=fx2(k,1)+
     +					 Resid(1)*Cheb(j1-1,zbar(i1,1))*
     +  	                 Cheb(j2-1,kbar(i2,1))*Cheb(j3-1,cbar(i3,1))*
     +                     dsqrt(1.0-(zbar(i1,1)**2))*
     +                     dsqrt(1.0-(kbar(i2,1)**2))*
     +                     dsqrt(1.0-(cbar(i3,1)**2))
	
                           fx2(k,2)=fx2(k,2)+
     +					 Resid(2)*Cheb(j1,zbar(i1,1))*
     +  	                 Cheb(j2,kbar(i2,1))*Cheb(j3,cbar(i3,1))*
     +                     dsqrt(1.0-(zbar(i1,1)**2))*
     +                     dsqrt(1.0-(kbar(i2,1)**2))*
     +                     dsqrt(1.0-(cbar(i3,1)**2))
	                     k=k+1
	                  End Do
                     End Do
	            End Do
	       End Do
	    End Do
      End Do
      fx(1:npar)=fx2(1:npar,1)
	fx(1+npar:2*npar)=fx2(1:npar,2)
      close(21)
	Return

	End Subroutine

	
C The subroutine GetResid returns the residual computed from the two Euler equations
C of the model. The functions that are approximated are q1 and the conditional
C expectation in the Euler equation for consumption


      Subroutine GetResid(z,k,c,x)

      use Def1
      use Errors

	Real(8) z, k, c, x(1:2)
      
      Integer(4) j
	Real(8) k2,c1,c2,q1,q2, lam1, lam2, x0, pvec(1:nz), i1, i2
	Real(8) GetPsi, GetInv, GetPhi, psi2, icrit


   
      x(1:2)=0.0

      q1=GetPsi(gvec(1:npar,1),z,k,c)
      if (IsNan(q1)) then
	  FError=1
	  x(1)=q1
	  return
	endif
      q1=beta*q1
      i1=GetInv(q1,k)      ! cannot be negative 
      c1=dexp(z)*(k**alpha)-i1
      if ((c1.lt.cmin) .or. (c1.gt.cmax)) then
	  FError=2
	  x(1)=0.0/0.0
	  return
	endif
      
	k2=GetPhi(i1,k)*k + (1-delta)*k   
      if ((k2.lt.kmin) .or. (k2.gt.kmax)) then
	 FError=3
	 x(1)=0.0/0.0
	 return
	endif

      psi2=GetPsi(gvec(1:npar,2),z,k,c)
      if (IsNan(psi2)) then
        FError=4
        x(1)=0.0/0.0
       return
      endif
      lam1=((c1-b*c)**(-eta)) - beta*b*psi2
      if (lam1.lt.0.0) then
        FError=5
        x(1)=0.0/0.0
       return
      endif
      
      Call GetProb(z,pvec)
      Do j=1,nz
            
       q2=GetPsi(gvec(1:npar,1),zgrid(j),k2,c1)
       if (IsNan(q2)) then
	   FError=6
	   x(1)=0.0/0.0
	   return
	  endif
	  q2=beta*q2
        i2=GetInv(q2,k2)
        c2=dexp(zgrid(j))*(k2**alpha) - i2
C        if ((c2.lt.cmin) .or. (c2.gt.cmax)) then
C	     FError=7
C	     x(1)=0.0/0.0
C	     return
C	  endif
        lam2=GetPsi(gvec(1:npar,2),zgrid(j),k2,c1)
        if (IsNan(lam2)) then
	    FError=8
		x(1)=0.0/0.0
		return
	  endif
	  lam2=((c2-b*c1)**(-eta))-beta*b*lam2
        if (lam2.lt.0.0) then
          FError=9
          x(1)=0.0/0.0
          return
        endif                
        x0=beta*(lam2/lam1)*( alpha*dexp(zgrid(j))*(k2**(alpha-1))
     +  - (i2/k2)+q2*(GetPhi(i2,k2)+1-delta) )
       x(1)=x(1)+pvec(j)*x0
       x0=((c2-b*c1)**(-eta))
       x(2)=x(2)+pvec(j)*x0
     
      end do

      x(1)=q1-x(1)
      x(2)=psi2-x(2)

      Return

	End Subroutine


C GetPsi: Computes Psi(gamma,z,k,c)

      Real(8) Function GetPsi(gamma,z,k,c)

      use Def1
	use Errors
	
	Real(8) gamma(1:npar), z, k, c

	Real(8) x(1:3), ChebEval4

      x(1)=z
	x(2)=dlog(k)
	x(3)=dlog(c)

	GetPsi=ChebEval4(npar,gamma,degree,x,bounds1)

	If (GetPsi .gt. 700.) then      
	   FError=1
	   GetPsi=0./0.	   
	   return
	else
	   GetPsi=dexp(GetPsi)
	   if (GetPsi .le. 0.0) then
	      FError=1
	      GetPsi=0./0.	      
	   endif
	endif      

      Return

	End Function

C GetInv: this function returns investment, given q and k

      Real(8) Function GetInv(q,k)

	use Def1
      Real(8) q, k

      GetInv=k*((q*a1)**(1./zeta)) 

	Return

	End Function

C GetPhi: this function returns Phi(i/k)

      Real(8) Function GetPhi(i,k)

	use Def1

	Real(8) i, k

      GetPhi=(a1/(1.-zeta))*((i/k)**(1.-zeta))+a2

	Return

	End Function


C The Function GetNpar returns the number of coefficients of a complete
C polynomial in three independent variables of degree degree.
C

      Integer(4) Function GetNpar(degree)

	Integer(4) degree

      Integer(4) i1, i2, i3, k, p
     
	p=degree+1
      k=1

	Do i1=1,p
	   do i2=1,(p+1-i1)
	      do i3=1,(p+2-i1-i2)
              k=k+1
            end do
	   end do
	end do

      GetNpar=k-1

	Return

	End Function


	Subroutine PrintSolution(gamma)

	Use Def1

	Real(8) gamma(1:2*npar)

	Integer(4) k, p, i1, i2, i3
	Real(8)  gvec1(1:npar), gvec2(1:npar)

	gvec1=gamma(1:npar)
	gvec2=gamma(npar+1:2*npar)
	      
      K=1
	p=degree+1
	Write(11,"(A)") 'Polynomial     Function 1        Function 2'
	Do i1=1,p
	   Do i2=1,(p+1-i1)
	      Do i3=1,(p+2-i1-i2)
	       write(11,"(I1,I1,I1,2X,2(F10.6,7X))") (i1-1),(i2-1),(i3-1),
     +		 gvec1(k), gvec2(k)
	         k=k+1
	      enddo
         enddo
      enddo

      return

	End Subroutine



C GetProb: Computes the transition probability from state x0 to state x[i].

      Subroutine GetProb(x0,p)

      use Def1
	use IMSLF90       

      Real(8), Intent(IN)   :: x0
	Real(8), Intent(OUT) :: p(1:nz)

      Integer(4) i
	Real(8) w, dnordf

	w=zgrid(2)-zgrid(1)
      
	p(1)=dnordf((zgrid(1)-rho*x0 + 0.5*w)/sigma)

      Do i=2,nz-1,1
	    
	   p(i)= dnordf((zgrid(i)-rho*x0 + 0.5*w)/sigma)
     +        -dnordf((zgrid(i)-rho*x0 - 0.5*w)/sigma)
		
	End Do

	p(nz)=1.0-Sum(p(1:nz-1))

      Return

      End Subroutine


C GetGamma: Uses the vector of coefficients of a polynomial of degree-1
C           to initialize the vector of coefficients of a polynomial
C           of degree

      Subroutine GetGamma(p,n1,n2,gin,gout)      

	Integer(4) n1, n2, p
	Real(8)  gin(n1), gout(n2)

      Integer(4) i1,i2,i3, p1, p2, k1,k2
	
	gout=0.0

	p1=p+1
      p2=p1+1
      k1=1
	k2=0
      Do i1=1,p2,1
         do i2=1,(p2+1-i1),1
            do i3=1,(p2+2-i1-i2),1         
             if ((i1+i2+i3-2).le.p1) then
		      k2=k2+1
			  gout(k1)=gin(k2)
		   endif            
             k1=k1+1            
            End Do
         End Do
      End Do

      Return

	End Subroutine


C Simulate: Simulates the model and returns the rates of return

      Subroutine Simulate(par,xs,hit,rss,sy)

      use DFLIB
      use Def1
	use Errors
      

      Integer(4) hit(1:2,1:2)
	Real(8) par(2*npar), xs(1:3), rss(1:3),sy

	Integer(4) t, j, zindex(1:nobs), s
	Real(8) i1, GetPsi, GetInv, GetPhi, x, q2,i2,c2,l2,k2
	Real(8) pvec(1:nz), rs(1:3), yt(1:nobs+1), StDev, HPL

      Type (rccoord) curpos		

      gvec(1:npar,1)=par(1:npar)
	gvec(1:npar,2)=par(1+npar:2*npar)
      
	rss=0.0
      hit=0
      sy=0.0

      HPL=1600.0

	Do s=1,nofs

      rs=0.0

      qt(1)=xs(1)  ! in the first period the economy is the steady state
      kt(1)=xs(2)
	kt(2)=xs(2)
	ct(1)=xs(3)
	yt(1)=dexp(zs)*(kt(1)**alpha)

	lt(1)=(1-beta*b)*(((1-b)*xs(3))**(-eta))

      Call MarkovTP(nobs,nz,pmat,floor(dble(nz/2)),zindex)
	Do t=1,nobs
	   zt(t)=zgrid(zindex(t))
      End Do

	zt(1)=zs      
      
C      open(21,file='Text.txt')
	Do t=1,nobs	   
	   qt(t+1)=beta*GetPsi(gvec(1:npar,1),zt(t+1),kt(t+1),ct(t))
              i1=GetInv(qt(t+1),kt(t+1))
	   yt(t+1)=dexp(zt(t+1))*(kt(t+1)**alpha)
         ct(t+1)=dexp(zt(t+1))*(kt(t+1)**alpha)-i1
	   if (ct(t+1).lt.cmin) hit(1,1)=hit(1,1)+1
	   if (ct(t+1).gt.cmax) hit(1,2)=hit(1,2)+1
	 kt(t+2)=GetPhi(i1,kt(t+1))*kt(t+1)+(1-delta)*kt(t+1)
	   if (kt(t+2).lt.kmin) hit(2,1)=hit(2,1)+1
	   if (kt(t+2).gt.kmax) hit(2,2)=hit(2,2)+1
C	Write(21,"('t= ',I4,2X,4(F10.5,3X),4(I3,2X))") t, zt(t), qt(t),
C     +   ct(t), kt(t), hit(1,1), hit(1,2), hit(2,1), hit(2,2)
	   lt(t+1)=GetPsi(gvec(1:npar,2),zt(t+1),kt(t+1),ct(t))
	   lt(t+1)=((ct(t+1)-b*ct(t))**(-eta))-beta*b*lt(t+1)	   
	   
	   Call GetProb(zt(t+1),pvec)
	   x=0.0
	   Do j=1,nz
	      q2=beta*GetPsi(gvec(1:npar,1),zgrid(j),kt(t+2),ct(t+1))
	      i2=GetInv(q2,kt(t+2))
	      c2=exp(zgrid(j))*(kt(t+2)**alpha)-i2	      
	      l2=GetPsi(gvec(1:npar,2),zgrid(j),kt(t+2),ct(t+1))
	      l2=((c2-b*ct(t+1))**(-eta))-beta*b*l2
	      x=x+pvec(j)*l2	      
         end do
	   rt(t,1)=lt(t)/(beta*x)
	   rt(t,1)=rt(t,1)-1.
	   rt(t,2)=alpha*dexp(zt(t+1))*(kt(t+1)**alpha)-i1+qt(t+1)*kt(t+2)
         rt(t,2)=rt(t,2)/(qt(t)*kt(t+1))
	   rt(t,2)=rt(t,2)-1.
	   rt(t,3)=rt(t,2)-rt(t,1)
	End Do
	
	rs(1)=dble(sum(rt(1:nobs,1))/nobs)
	rs(2)=dble(sum(rt(1:nobs,2))/nobs)
	rs(3)=dble(sum(rt(1:nobs,3))/nobs)
      
	rss(1)=rss(1)+rs(1)
	rss(2)=rss(2)+rs(2)
	rss(3)=rss(3)+rs(3)

	CALL HPFilter(nobs+1,dlog(yt),HPL,lt)
	sy=sy+StDev(nobs+1,lt)
      
      End Do

	rss(1)=rss(1)/dble(nofs)
	rss(2)=rss(2)/dble(nofs)
	rss(3)=rss(3)/dble(nofs)
      sy=sy/dble(nofs)
C      close(21)
	Return
	
	End Subroutine  

C Cheb(p,x) : Returns the p-th degree of a Chebyshev-polynomial, x^p, x must be in [-1,1]
C             p=1 is the constant, p=2 is x, and so forth

      Real(8) Function Cheb(p,x)

	Integer(4) p
	Real(8) x

	Integer(4) i,p1
	Real(8)  x0,x1,x2


      Select Case(p)

        Case(0)
	     Cheb=1.0
	     Return

	  Case(1)
	     Cheb=x
	     Return

        Case(2:)
          x0=1.0	
	    x1=x	
          p1=p+1
	    do i=3,p1
	       x2=2.*x*x1-x0
	       x1=x2
	       x0=x1
          End Do	
	    Cheb=x2
      End Select

	Return

	End Function
	
      Subroutine Impulse(n,par,i,xs,tmax)

      use Def1

	Integer(4) i, n, tmax
	Real(8) par(1:n), xs(3)
	
	Integer(4) t
	Real(8) it(1:tmax), GetInv, GetPhi, GetPsi

      open(22,file='Impulse.txt')
	gvec(1:npar,1)=par(1:npar)
	gvec(1:npar,2)=par(1+npar:2*npar)

      qt(1)=xs(1)
      ct(1)=xs(3)
	kt(1)=xs(2)
	kt(2)=xs(2)
	zt(1)=zs
	zt(2)=zgrid(i)

	Do t=2,tmax
	   if (kt(t).lt.kmin) then
	     write(22,"('t= ',I3,'k<kmin')") t
		 goto 31
	   endif
	   if (kt(t).gt.kmax) then
	      write(22,"('t= ',I3,'k>kmin')") t
	      goto 31
	   endif
	   if (ct(t-1).lt.cmin) then
	     write(22,"('t= ',I3,'c<cmin')") t
	     goto 31
	   endif
	   if (ct(t-1).gt.cmax) then
	      write(22,"('t= ',I3,'c>cmin')") t
	      goto 31
	   endif
	   qt(t)=beta*GetPsi(gvec(1:npar,1),zt(t),kt(t),ct(t-1))
	   it(t)=GetInv(qt(t),kt(t))
	   ct(t)=exp(zt(t))*(kt(t)**alpha)-it(t)
	 kt(t+1)=GetPhi(it(t),kt(t))*kt(t)+(1-delta)*kt(t)
	 zt(t+1)=rho*zt(t)
	 Write(22,"('t= ',I4,2X,5(F10.5,2X))") t, zt(t), qt(t), it(t),
     + ct(t), kt(t)
     
      End Do
	
31	close(22)
	Return
	End Subroutine

	     
      Subroutine Writeqck(Gamma)

	use Def1
      use IMSLF90	
      use Errors
	
	Real(8) Gamma(1:2*Npar)

	Integer(4) i, i1, i2, i3, p1
	Real(8) q1,k1,inv1,c1,zbar(degree+1),kbar(degree+1), cbar(degree+1),
     +         pi, Dconst, GetPhi, GetInv, GetPsi

	pi=Dconst('pi')    
      p1=degree+1
	
	Do i=1,p1
	   zbar(i)=dcos(((dble(2.0*i)-1.0)/(dble(2.*p1))) *pi)
	   zbar(i)=0.5*(zbar(i)+1.0)*(zmax-zmin) + zmin
	End Do
	
	Do i=1,p1
	   kbar(i)=dcos(((dble(2.0*i)-1.0)/(dble(2.*p1)))*pi)   
	   kbar(i)=0.5*(kbar(i)+1.0)*(kmax-kmin) + kmin
	End Do

      Do i=1,p1
	   cbar(i)=dcos(((dble(2.0*i)-1.0)/(dble(2.*p1)))*pi)
	   cbar(i)=0.5*(cbar(i)+1.0)*(cmax-cmin) + cmin
	End Do      

      gvec(1:npar,1)=Gamma(1:npar)
	gvec(1:npar,2)=Gamma(1+npar:2*npar)      

	Do i1=1,p1
	     do i2=1,p1+1-i1
	          do i3=1,p1+2-i1-i2
	      q1=beta*GetPsi(gvec(1:npar,1),zbar(i1),kbar(i2),cbar(i3))         
            inv1=GetInv(q1,kbar(i2))
            c1=dexp(zbar(i1))*(kbar(i2)**alpha)-inv1
	      k1=GetPhi(inv1,kbar(i2))*kbar(i2)+(1-delta)*kbar(i2)
	      Write(11,"(6(F10.5,2X))") zbar(i1),kbar(i2),cbar(i3),
     +		  q1,(c1/cs),(k1/ks)
	End Do
	End Do
	End Do

	Return

	End Subroutine

       